/// <reference path="../../../typings/index.d.ts" />
"use strict";
const path = require("path");
const fs = require("fs");
const async = require("async");
const DEFAULT_BUFFER_SIZE = 8096;
exports.staticMineTypes = {
    ".png": {
        "MINE": "image/png",
        "cachable": false
    },
    ".gif": {
        "MINE": "image/gif",
        "cachable": false
    },
    ".jpg": {
        "MINE": "image/jpeg",
        "cachable": false
    },
    ".jpeg": {
        "MINE": "image/jpeg",
        "cachable": false
    },
    ".bmp": {
        "MINE": "image/bmp",
        "cachable": false
    },
    ".svg": {
        "MINE": "image/svg+xml",
        "cachable": false
    },
    ".ico": {
        "MINE": "image/x-icon",
        "cachable": false
    },
    ".tar": {
        "MINE": "application/x-tar",
        "cachable": false
    },
    ".tgz": {
        "MINE": "application/x-compressed",
        "cachable": false
    },
    ".txt": {
        "MINE": "text/plain",
        "cachable": false
    },
    ".json": {
        "MINE": "application/json",
        "cachable": true
    },
    ".zip": {
        "MINE": "application/zip",
        "cachable": false
    },
    ".html": {
        "MINE": "text/html",
        "cachable": true
    },
    ".htm": {
        "MINE": "text/html",
        "cachable": true
    },
    ".css": {
        "MINE": "text/css",
        "cachable": true
    },
    ".js": {
        "MINE": "application/x-javascript",
        "cachable": true
    },
    "default": {
        "MINE": "application/octet-stream",
        "cachable": false
    }
};
function StaticServer(opts) {
    if (!opts.prefix.endsWith("/")) {
        opts.prefix += "/";
    }
    if (!opts.root.endsWith("/")) {
        opts.root += "/";
    }
    opts.bufferSize || (opts.bufferSize = DEFAULT_BUFFER_SIZE);
    return function (req, resp, next, server) {
        if (!req.path.startsWith(opts.prefix)) {
            return next(null, false);
        }
        let fileName = "./" + path.simplify(opts.root + req.path.substr(opts.prefix.length));
        if (!fileName) {
            return next({
                "name": "NOT-FOUND",
                "message": "Specific resource is not found!"
            });
        }
        let mine = exports.staticMineTypes[path.extname(fileName).toLowerCase()] || exports.staticMineTypes["default"];
        let fileSize;
        let filePointer;
        async.series([
            function (next) {
                fs.exists(fileName, function (exist) {
                    if (!exist) {
                        return next({
                            "name": "NOT-FOUND",
                            "message": "Specific resource is not found!"
                        });
                    }
                    next();
                });
            },
            function (next) {
                fs.stat(fileName, function (err, stat) {
                    if (err || stat.isDirectory()) {
                        return next({
                            "name": "NOT-FOUND",
                            "message": "Specific resource is not found!"
                        });
                    }
                    fileSize = stat.size;
                    next();
                });
            },
            function (next) {
                fs.open(fileName, "r+", function (err, fd) {
                    if (err) {
                        return next({
                            "name": "FORBIDDEN",
                            "message": "Specific resource is not accessible!"
                        });
                    }
                    filePointer = fd;
                    next();
                });
            },
            function (next) {
                let bytesRead = 0;
                resp.writeHead(200, {
                    "Content-Length": fileSize,
                    "Content-Type": mine.MINE
                });
                async.whilst(function () {
                    return bytesRead < fileSize;
                }, function (next) {
                    let buf = new Buffer(opts.bufferSize);
                    fs.read(filePointer, buf, 0, opts.bufferSize, null, (err, bytes, bf) => {
                        if (err) {
                            return next({
                                "name": "INTERNAL",
                                "message": "Failed to read file."
                            });
                        }
                        if (bytes > 0) {
                            bytesRead += bytes;
                            resp.write(buf, function () {
                                next();
                            });
                        }
                        else {
                            next();
                        }
                    });
                }, function (err) {
                    resp.end();
                    fs.close(filePointer, next);
                });
            }
        ], function (err) {
            next(err);
        });
    };
}
exports.StaticServer = StaticServer;
//# sourceMappingURL=static-server.js.map