/// <reference path="../../../typings/index.d.ts" />

import * as HTTPServer from "../server";
import path = require("path");
import fs = require("fs");
import async = require("async");

export interface MINEConfig {
    "MINE": string;
    "cachable": boolean;
    "maxCacheSize"?: number;
}

const DEFAULT_BUFFER_SIZE: number = 8096;

export let staticMineTypes: LRTS.types.HashMap<MINEConfig> = {
    ".png": {
        "MINE": "image/png",
        "cachable": false
    },
    ".gif": {
        "MINE": "image/gif",
        "cachable": false
    },
    ".jpg": {
        "MINE": "image/jpeg",
        "cachable": false
    },
    ".jpeg": {
        "MINE": "image/jpeg",
        "cachable": false
    },
    ".bmp": {
        "MINE": "image/bmp",
        "cachable": false
    },
    ".svg": {
        "MINE": "image/svg+xml",
        "cachable": false
    },
    ".ico": {
        "MINE": "image/x-icon",
        "cachable": false
    },
    ".tar": {
        "MINE": "application/x-tar",
        "cachable": false
    },
    ".tgz": {
        "MINE": "application/x-compressed",
        "cachable": false
    },
    ".txt": {
        "MINE": "text/plain",
        "cachable": false
    },
    ".json": {
        "MINE": "application/json",
        "cachable": true
    },
    ".zip": {
        "MINE": "application/zip",
        "cachable": false
    },
    ".html": {
        "MINE": "text/html",
        "cachable": true
    },
    ".htm": {
        "MINE": "text/html",
        "cachable": true
    },
    ".css": {
        "MINE": "text/css",
        "cachable": true
    },
    ".js": {
        "MINE": "application/x-javascript",
        "cachable": true
    },
    "default": {
        "MINE": "application/octet-stream",
        "cachable": false
    }
};

export interface Options {

    prefix: string;

    root: string;

    bufferSize?: number;
}

export function StaticServer(opts: Options): HTTPServer.Middleware {

    if (!opts.prefix.endsWith("/")) {
        opts.prefix += "/";
    }

    if (!opts.root.endsWith("/")) {
        opts.root += "/";
    }

    opts.bufferSize || (opts.bufferSize = DEFAULT_BUFFER_SIZE);

    return function(
        req: HTTPServer.Request,
        resp: HTTPServer.Response,
        next: HTTPServer.MiddlewareNextCallback,
        server: HTTPServer.ServerPrototype
    ): void {

        if (!req.path.startsWith(opts.prefix)) {

            return next(null, false);

        }

        let fileName: string = "./" + path.simplify(opts.root + req.path.substr(opts.prefix.length));

        if (!fileName) {

            return next({
                "name": "NOT-FOUND",
                "message": "Specific resource is not found!"
            });
        }

        let mine: MINEConfig = staticMineTypes[path.extname(fileName).toLowerCase()] || staticMineTypes["default"];

        let fileSize: number;
        let filePointer: number;

        async.series([

            function (next: ErrorCallback) {

                fs.exists(fileName, function(exist: boolean): void {

                    if (!exist) {

                        return next({
                            "name": "NOT-FOUND",
                            "message": "Specific resource is not found!"
                        });
                    }

                    next();
                });
            },

            function (next: ErrorCallback) {

                fs.stat(fileName, function(err: Error, stat: fs.Stats): void {

                    if (err || stat.isDirectory()) {

                        return next({
                            "name": "NOT-FOUND",
                            "message": "Specific resource is not found!"
                        });
                    }

                    fileSize = stat.size;

                    next();
                });

            },

            function (next: ErrorCallback) {

                fs.open(fileName, "r+", function(err: Error, fd: number): void {

                    if (err) {

                        return next({
                            "name": "FORBIDDEN",
                            "message": "Specific resource is not accessible!"
                        });
                    }

                    filePointer = fd;

                    next();

                });

            },

            function (next: ErrorCallback) {

                let bytesRead: number = 0;

                resp.writeHead(200, {
                    "Content-Length": fileSize,
                    "Content-Type": mine.MINE
                });

                async.whilst(function(): boolean {

                    return bytesRead < fileSize;

                }, function(next: ErrorCallback): void {

                    let buf: Buffer = new Buffer(opts.bufferSize);

                    fs.read(filePointer, buf, 0, opts.bufferSize, null,
                        (err: Error, bytes: number, bf: Buffer): void => {

                        if (err) {

                            return next({
                                "name": "INTERNAL",
                                "message": "Failed to read file."
                            });
                        }

                        if (bytes > 0) {

                            bytesRead += bytes;

                            resp.write(buf, function() {
                                next();
                            });

                        } else {

                            next();
                        }

                    });

                }, function(err?: Error): void {

                    resp.end();
                    fs.close(filePointer, next);

                });

            }

        ], function (err?: Error) {

            next(err);
        });

    };
}
