/// <reference path="../../typings/index.d.ts" />

import HTTPServer = LRTS.http.server;
import callbacks = LRTS.globals.callbacks;
import NodeHTTP = require("http");
import NodeURL = require("url");
import async = require("async");
import { EmptyCallback } from "../utils/index";

export enum Status {
    "IDLE", "STARTING", "WORKING", "CLOSING", "CLOSED", "ERROR"
}

export enum StatusCode {
    "CONTINUE" = 100,
    "SWITCHING_PROTOCOLS" = 101,
    "PROCESSING" = 102,
    "OK" = 200,
    "CREATED" = 201,
    "ACCEPTED" = 202,
    "NON_AUTHORITATIVE_INFORMATION" = 203,
    "NO_CONTENT" = 204,
    "RESET_CONTENT" = 205,
    "PARTIAL_CONTENT" = 206,
    "MULTIPLE_CHOICES" = 300,
    "MOVED_PERMANENTLY" = 301,
    "MOVE_TEMPORARILY" = 302,
    "SEE_OTHER" = 303,
    "NOT_MODIFIED" = 304,
    "USE_PROXY" = 305,
    "SWITCH_PROXY" = 306,
    "TEMPORARY_REDIRECT" = 307,
    "BAD_REQUEST" = 400,
    "UNAUTHORIZED" = 401,
    "FORBIDDEN" = 403,
    "NOT_FOUND" = 404,
    "METHOD_NOT_ALLOWED" = 405,
    "NOT_ACCEPTABLE" = 406,
    "PROXY_AUTHENTICATION_REQUIRED" = 407,
    "REQUEST_TIMEOUT" = 408,
    "CONFLICT" = 409,
    "GONE" = 410,
    "LENGTH_REQUIRED" = 411,
    "PRECONDITION_FAILED" = 412,
    "REQUEST_ENTITY_TOO_LARGE" = 413,
    "REQUEST_URI_TOO_LONG" = 414,
    "UNSUPPORTED_MEDIA_TYPE" = 415,
    "REQUESTED_RANGE_NOT_SATISFIABLE" = 416,
    "EXPECTATION_FAILED" = 417,
    "UNPROCESSABLE_ENTITY" = 422,
    "LOCKED" = 423,
    "FAILED_DEPENDENCY" = 424,
    "UNORDERED_COLLECTION" = 425,
    "UPGRADE_REQUIRED" = 426,
    "RETRY_WITH" = 449,
    "INTERNAL_SERVER_ERROR" = 500,
    "NOT_IMPLEMENTED" = 501,
    "BAD_GATEWAY" = 502,
    "SERVICE_UNAVAILABLE" = 503,
    "GATEWAY_TIMEOUT" = 504,
    "HTTP_VERSION_NOT_SUPPORTED" = 505,
    "VARIANT_ALSO_NEGOTIATES" = 506,
    "INSUFFICIENT_STORAGE" = 507,
    "BANDWIDTH_LIMIT_EXCEEDED" = 509,
    "NOT_EXTENDED" = 510,
    "UNPARSEABLE_RESPONSE_HEADERS" = 600
}

export interface Request extends NodeHTTP.IncomingMessage {

    queryString: string;

    query: LRTS.types.HashMap<any>;

    requestPath: string;

    path: string;

    params: LRTS.types.HashMap<any>;

}

export interface RouterHandler {

    (request: Request, response: NodeHTTP.ServerResponse, next: ErrorCallback, server: ServerPrototype): void;
}

interface RegExpRouterRule {

    "exp": RegExp;

    "fields"?: string[];

    "handler": RouterHandler;
}

interface MiddlewareRule {

    "regExp"?: RegExp;

    "strExp"?: string;

    "handler": Middleware;
}

export interface MiddlewareNextCallback {

    (err: Error, interrupted?: boolean): void;
}

export interface Middleware {

    (request: Request, response: NodeHTTP.ServerResponse, next: MiddlewareNextCallback, server: ServerPrototype): void;
}

export interface ServerPrototype {

    on(event: string, handler: RouterHandler): ServerPrototype;

    all(routeRule: HTTPServer.RouteRule, handler: RouterHandler): ServerPrototype;

    before(routeRule: any, handler: Middleware): ServerPrototype;

    get(routeRule: HTTPServer.RouteRule, handler: RouterHandler): ServerPrototype;
    patch(routeRule: HTTPServer.RouteRule, handler: RouterHandler): ServerPrototype;
    post(routeRule: HTTPServer.RouteRule, handler: RouterHandler): ServerPrototype;
    options(routeRule: HTTPServer.RouteRule, handler: RouterHandler): ServerPrototype;
    head(routeRule: HTTPServer.RouteRule, handler: RouterHandler): ServerPrototype;
    put(routeRule: HTTPServer.RouteRule, handler: RouterHandler): ServerPrototype;
    delete(routeRule: HTTPServer.RouteRule, handler: RouterHandler): ServerPrototype;
    trace(routeRule: HTTPServer.RouteRule, handler: RouterHandler): ServerPrototype;
    connect(routeRule: HTTPServer.RouteRule, handler: RouterHandler): ServerPrototype;
}

const DEFAULT_HOST: string     = "127.0.0.1";
const DEFAULT_PORT: number     = 80;
const DEFAULT_BACKLOG: number  = 511;

export interface Response extends NodeHTTP.ServerResponse {

}

export class Server implements ServerPrototype {

    protected serverStatus: Status;

    protected serverOptions: HTTPServer.Options;

    protected rootServer: Server;

    protected rawURIRouters: LRTS.types.HashMap<LRTS.types.HashMap<RouterHandler>>;

    protected regexpURIRouters: LRTS.types.HashMap<RegExpRouterRule[]>;

    protected middlewares: MiddlewareRule[];

    protected httpServer: NodeHTTP.Server;

    protected eventHandlers: LRTS.types.HashMap<RouterHandler[]>;

    protected static defaultEventHandlers: LRTS.types.HashMap<RouterHandler>;

    public on(event: string, handler: RouterHandler): Server {

        if (undefined === this.eventHandlers[event]) {

            this.eventHandlers[event] = [];
        }

        this.eventHandlers[event].push(handler);

        return this;
    }

    public get backlog(): number {

        return this.serverOptions.backlog;
    }

    public set backlog(newValue: number) {

        if (this.serverStatus !== Status.IDLE) {

            throw {
                "name": "BAD-STATUS",
                "message": "You can't change backlog while server is working."
            };
        }

        this.serverOptions.backlog = newValue;
    }

    public get port(): number {

        return this.serverOptions.port;
    }

    public set port(newValue: number) {

        if (this.serverStatus !== Status.IDLE) {

            throw {
                "name": "BAD-STATUS",
                "message": "You can't change port while server is working."
            };
        }

        this.serverOptions.port = newValue;
    }

    public get host(): string {

        return this.serverOptions.host;
    }

    public set host(newValue: string) {

        if (this.serverStatus !== Status.IDLE) {

            throw <Error>{
                "name": "BAD-STATUS",
                "message": "You can't change host while server is working."
            };
        }

        this.serverOptions.host = newValue;
    }

    public get status(): Status {

        return this.serverStatus;
    }

    protected registerStringRouter(method: string, routeRule: string, handler: RouterHandler): void {

        if (routeRule.endsWith("/") && routeRule.length > 1) {

            routeRule = routeRule.substr(0, routeRule.length - 1);

        }

        console.debug("Router: Added String-Router =", routeRule);

        this.rawURIRouters[method][routeRule] = handler;
    }

    protected registerStrExpRouter(method: string, routeRule: string, handler: RouterHandler): void {

        if (routeRule.endsWith("/") && routeRule.length > 1) {

            routeRule = routeRule.substr(0, routeRule.length - 1);

        }

        console.debug("Router: Added Smart-Router =", routeRule);

        let fields: string[] = [];

        routeRule = routeRule.replace(/\<.+?\>/g, function(el: string): string {

            let rtn: string = "(.+?)", pos: number = el.indexOf(":");

            if (pos >= 0) {

                fields.push(el.substr(1, pos - 1));
                rtn = "(" + el.substring(pos + 1, el.length - 1) + ")";
            }
            else {

                fields.push(el.substr(1, el.length - 2));
            }

            return rtn;
        });

        this.regexpURIRouters[method].push({

            "exp": new RegExp("^" + routeRule + "$"),

            "fields": fields,

            "handler": handler
        });

    }

    /**
     * This method register a new handler for a routing-rule.
     */
    protected registerRouteHandler(method: string, routeRule: HTTPServer.RouteRule, handler: RouterHandler): void {

        if (typeof routeRule === "string") {

            if (routeRule.match(/\<.+?\>/)) {

                this.registerStrExpRouter(method, routeRule, handler);

            } else {

                this.registerStringRouter(method, routeRule, handler);
            }


        } else if (routeRule instanceof RegExp) {

            console.debug("Router: Added Smart-Router =", routeRule);

            this.regexpURIRouters[method].push({

                "exp": routeRule,

                "handler": handler
            });

        }
    }

    public all(rule: HTTPServer.RouteRule, handler: RouterHandler): Server {

        this.registerRouteHandler("ALL", rule, handler);

        return this;
    }

    public get(rule: HTTPServer.RouteRule, handler: RouterHandler): Server {

        this.registerRouteHandler("GET", rule, handler);

        return this;
    }

    public patch(rule: HTTPServer.RouteRule, handler: RouterHandler): Server {

        this.registerRouteHandler("PATCH", rule, handler);

        return this;
    }

    public post(rule: HTTPServer.RouteRule, handler: RouterHandler): Server {

        this.registerRouteHandler("POST", rule, handler);

        return this;
    }

    public options(rule: HTTPServer.RouteRule, handler: RouterHandler): Server {

        this.registerRouteHandler("OPTIONS", rule, handler);

        return this;
    }

    public head(rule: HTTPServer.RouteRule, handler: RouterHandler): Server {

        this.registerRouteHandler("HEAD", rule, handler);

        return this;
    }

    public put(rule: HTTPServer.RouteRule, handler: RouterHandler): Server {

        this.registerRouteHandler("PUT", rule, handler);

        return this;
    }

    public delete(rule: HTTPServer.RouteRule, handler: RouterHandler): Server {

        this.registerRouteHandler("DELETE", rule, handler);

        return this;
    }

    public trace(rule: HTTPServer.RouteRule, handler: RouterHandler): Server {

        this.registerRouteHandler("TRACE", rule, handler);

        return this;
    }

    public connect(rule: HTTPServer.RouteRule, handler: RouterHandler): Server {

        this.registerRouteHandler("CONNECT", rule, handler);

        return this;
    }

    public constructor(opts?: HTTPServer.Options) {

        this.serverStatus = Status.IDLE;

        this.serverOptions = {

            "host": opts !== void 0 && opts.host !== void 0 ? opts.host : DEFAULT_HOST,
            "port": opts !== void 0 && opts.port !== void 0 ? opts.port : DEFAULT_PORT,
            "backlog": opts !== void 0 && opts.backlog !== void 0 ? opts.backlog : DEFAULT_BACKLOG
        };

        this.middlewares = [];

        this.eventHandlers = {};

        this.rawURIRouters = {
            "GET": {},
            "POST": {},
            "PUT": {},
            "PATCH": {},
            "DELETE": {},
            "OPTIONS": {},
            "HEAD": {},
            "TRACE": {},
            "CONNECT": {},
            "ALL": {}
        };

        this.regexpURIRouters = {
            "GET": [],
            "POST": [],
            "PUT": [],
            "PATCH": [],
            "DELETE": [],
            "OPTIONS": [],
            "HEAD": [],
            "TRACE": [],
            "CONNECT": [],
            "ALL": []
        };

        for (let ev in Server.defaultEventHandlers) {

            this.eventHandlers[ev] = [Server.defaultEventHandlers[ev]];
        }

    }

    public before(exp: any, middleware?: Middleware): Server {

        if (exp instanceof RegExp && middleware) {

            this.middlewares.push({
                "regExp": exp,
                "handler": middleware
            });

        } else if (typeof exp === "string" && middleware) {

            this.middlewares.push({
                "strExp": exp,
                "handler": middleware
            });

        } else {

            this.middlewares.push({
                "handler": exp
            });
        }

        return this;
    }

    public start(cb?: ErrorCallback): Server {

        if (this.serverStatus !== Status.IDLE) {

            cb({
                "name": "BAD-OPERATION",
                "message": "HTTP Server has been already started."
            });

            return this;
        }

        let self: Server = this;

        this.serverStatus = Status.STARTING;

        this.httpServer = NodeHTTP.createServer(this.gateway.bind(this));

        this.httpServer.once("error", function(err: NodeJS.ErrnoException) {

            if (err && err.syscall === "listen") {

                cb && cb({
                    "name": "LISTEN-FAILURE",
                    "message": `Failed to listen on ${self.host}:${self.port}.`
                });
            }
        });

        this.httpServer.listen(this.serverOptions.port, this.serverOptions.host, this.serverOptions.backlog, function() {

            self.serverStatus = Status.WORKING;

            self.httpServer.emit("error");

            cb && cb();

        });

        return this;
    }

    protected gateway(req: Request, resp: NodeHTTP.ServerResponse): void {

        let server: Server = this;

        if (req.path === undefined) {

            let url: NodeURL.Url = NodeURL.parse(req.url);

            if (url.pathname.length > 1 && url.pathname[url.pathname.length - 1] === "/") {

                req.requestPath = url.pathname.substr(0, url.pathname.length - 1);

            } else {

                req.requestPath = url.pathname;
            }

            req.queryString = url.query;

            url = undefined;

            req.path = req.requestPath;

        }

        req.params = {};

        async.series([

            function(next: ErrorCallback): void {

                async.forEachOf<MiddlewareRule>(server.middlewares, function(
                    item: MiddlewareRule,
                    key: any,
                    next: ErrorCallback
                ): void {

                    if (item.strExp && item.strExp === req.path) {
                        setImmediate(item.handler, req, resp,
                        (err: Error, interrupted?: boolean): void => {

                            if (err) {

                                return next(err);
                            }

                            next();

                        }, server);
                    } else if (item.strExp === undefined && item.regExp === undefined) {

                        setImmediate(item.handler, req, resp,
                        (err: Error, interrupted?: boolean): void => {

                            if (err) {

                                return next(err);
                            }

                            next();

                        }, server);
                    }

                }, next);

            },

            function(next: ErrorCallback): void {

                let handler: RouterHandler;

                if (server.rawURIRouters[req.method] === undefined) {

                    setImmediate(server.eventHandlers["BAD-METHOD"], req, resp, next, server.rootServer ? server.rootServer : this);

                    return;
                }

                /**
                 * Lookup Text-URI handlers of specific method.
                 */
                handler = server.rawURIRouters[req.method][req.path];

                /**
                 * Lookup RegExp-URI handlers of specific method.
                 */
                if (!handler) {

                    let rules: RegExpRouterRule[] = server.regexpURIRouters[req.method];

                    for (let rule of rules) {

                        let matches: RegExpMatchArray;

                        if (matches = req.path.match(rule.exp)) {

                            if (rule.fields) {

                                rule.fields.forEach(function(el: string, index: number) {

                                    req.params[el] = matches[index + 1];
                                });
                            }

                            handler = rule.handler;
                            break;
                        }
                    }

                }

                /**
                 * Lookup Text-URI handlers of any method.
                 */
                if (!handler) {

                    handler = server.rawURIRouters["ALL"][req.path];

                }

                /**
                 * Lookup RegExp-URI handlers of any method.
                 */
                if (!handler) {

                    let rules: RegExpRouterRule[] = server.regexpURIRouters["ALL"];

                    for (let rule of rules) {

                        let matches: RegExpMatchArray;

                        if (matches = req.path.match(rule.exp)) {

                            if (rule.fields) {

                                rule.fields.forEach(function(el: string, index: number) {

                                    req.params[el] = matches[index + 1];
                                });
                            }

                            handler = rule.handler;
                            break;
                        }
                    }

                }

                if (handler) {

                    setImmediate(handler, req, resp, next, server.rootServer ? server.rootServer : this);

                } else {

                    next({
                        "name": "NOT-FOUND",
                        "message": null
                    });

                }

            }

        ], function(err?: Error): void {

            server.trigger(err ? err.name : "COMPLETED", req, resp);

        });

    }

    protected trigger(eventName: string, req: Request, resp: Response): Server {

        if (this.eventHandlers[eventName]) {

            let server: Server = this;

            async.forEachOf<RouterHandler>(this.eventHandlers[eventName], function(
                eh: RouterHandler,
                key: any,
                next: ErrorCallback
            ): void {

                eh(req, resp, next, server);

            }, EmptyCallback);

        } else {

            Server.defaultEventHandlers[eventName] && setImmediate(
                Server.defaultEventHandlers[eventName],
                req, resp, EmptyCallback
            );
        }

        return this;
    }

    public static init() {

        this.defaultEventHandlers = {};

        for (let eh of [
            {  "id": "NOT-FOUND",   "code": StatusCode.NOT_FOUND,              "message": "NOT FOUND"},
            {  "id": "BAD-METHOD",  "code": StatusCode.METHOD_NOT_ALLOWED,     "message": "METHOD NOT ALLOWED"},
            {  "id": "FORBIDDEN",   "code": StatusCode.FORBIDDEN,              "message": "FORBIDDEN"},
            {  "id": "INTERNAL",    "code": StatusCode.INTERNAL_SERVER_ERROR,  "message": "INTERNAL ERROR"}
        ]) {

            this.defaultEventHandlers[eh.id] = (request: Request, response: NodeHTTP.ServerResponse, next: ErrorCallback, server: ServerPrototype): void => {

                response.statusCode = eh.code;
                response.statusMessage = eh.message;
                response.end(`HTTP ${eh.code}: ${eh.message}`);

                next();
            };

        }

        delete this.init;
    }

}

Server.init && Server.init();
